/***************************************************************************
 *   Copyright (C) 2007 by Antony Martin                                   *
 *   martin.antony@yahoo.fr                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Cree le : 11 mars 2007
   derniere modification le 18/03/2007 */


#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEMath.h>
#include <SCE/utils/SCEString.h>

#include <SCE/core/SCECSupport.h>
#include <SCE/core/SCECTexture.h>

#include <SCE/interface/SCEDDSLoader.h>
#include <SCE/interface/SCEMedia.h>
#include <SCE/interface/SCETexture.h>


static unsigned int sce_tex_reduce; /* booleen: reduction des images ? */
static SCEenum sce_tex_filter;      /* filtre de redimensionnement */


int SCE_Texture_MediaTypeID(void)
{
    static int id = -1;
    
    if(id == -1)
        id = SCE_Media_GenTypeID();
    
    return (id);
}


int SCE_Texture_Init(void)
{
    /* 'magic' du dds */
    const char *dds_magic_str = "DDS ";
    int *dds_magic = (int*)dds_magic_str;
    
    /* initialisation de devil */
    ilInit();
    iluInit();
    
    ilOriginFunc(IL_ORIGIN_LOWER_LEFT);
    ilEnable(IL_ORIGIN_SET);
    
    /* envoie des loaders au media manager */
    SCE_Media_RegisterLoader(SCE_Texture_MediaTypeID(), 0, ".png .gif .jpg "
                             ".jpeg .bmp .ico .mng .pcx .rgb .rgba .tga .tif",
                             SCE_Texture_DevILLoader);
    
    SCE_Media_RegisterLoader(SCE_Texture_MediaTypeID(), *dds_magic, ".dds",
                             SCE_Texture_DDSLoader);
    
    /* par defaut, on reduit les images lors du redimensionnement */
    sce_tex_reduce = 1;
    sce_tex_filter = SCE_FILTER_NEAREST;
    
    return (SCE_OK);
}

int SCE_Texture_Quit(void)
{
    /* fermeture de devil */
    ilShutDown();
    
    return (SCE_OK);
}



/* cree une nouvelle structure de donnees d'image */
static SCE_STexData* SCE_Texture_CreateData(void)
{
    SCE_STexData *data = NULL;
    
    data = SCE_malloc(sizeof *data);
    if(data == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    data->cdata = SCE_malloc(sizeof *data->cdata);
    if(data->cdata == NULL)
    {
        SCE_free(data);
        Logger_LogSrc();
        return (NULL);
    }
    
    data->id = 0;
    data->n_data = 1;
    data->freetexels = 0;
    
    return (data);
}

/* detruit une donnee image */
static void SCE_Texture_DeleteData(SCE_STexData *data)
{
    if(data != NULL)
    {
        if(ilIsImage(data->id))
            ilDeleteImages(1, &data->id);
        
        if(data->n_data > 0 && data->cdata != NULL)
        {
            /* les suivants sont forcement des niveaux de mipmap et par la
               meme occasion, viennent du meme pointeur, on libere donc
               uniquement le premier */
            if(data->freetexels)
                SCE_free(data->cdata[0].texels);
            SCE_free(data->cdata);
        }
        
        SCE_free(data);
    }
}


SCE_STexture* SCE_Texture_Create(SCEenum type)
{
    SCE_STexture *tex = NULL;
    
    tex = SCE_malloc(sizeof *tex);
    if(tex == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    tex->tex = SCE_CCreateTexture(type);
    if(tex->tex == NULL)
    {
        SCE_free(tex);
        Logger_LogSrc();
        return (NULL);
    }
    
    tex->n_data = 0;
    tex->data = NULL;
    
    tex->loaded = tex->builded = 0;
    
    return (tex);
}

void SCE_Texture_Delete(SCE_STexture *tex)
{
    unsigned int i;
    
    if(tex != NULL)
    {
        for(i=0; i<tex->n_data; i++)
            SCE_Texture_DeleteData(tex->data[i]);
        
        SCE_CDeleteTexture(tex->tex);
        SCE_free(tex);
    }
}


#if 0
SCEenum SCE_Texture_GetImageType(const char *ext)
{
    SCEenum type = 0;
    
    /* macro tres pratique */
    #define SCE_TEX_RESEARCHFOR(extension, constante)\
    if(strcmp(ext, extension) == 0)\
        type = constante;\
    else
    
    SCE_TEX_RESEARCHFOR(".bmp", IL_BMP)
    SCE_TEX_RESEARCHFOR(".cut", IL_CUT)
    SCE_TEX_RESEARCHFOR(".dcx", IL_DCX)
    SCE_TEX_RESEARCHFOR(".dds", IL_DDS)/*
    SCE_TEX_RESEARCHFOR(".doom", IL_DOOM)
    SCE_TEX_RESEARCHFOR(".doomfloat", IL_DOOM_FLOAT)*/
    SCE_TEX_RESEARCHFOR(".gif", IL_GIF)
    SCE_TEX_RESEARCHFOR(".hdr", IL_HDR)
    SCE_TEX_RESEARCHFOR(".ico", IL_ICO)
    SCE_TEX_RESEARCHFOR(".jpg", IL_JPG)
    SCE_TEX_RESEARCHFOR(".jpeg", IL_JPG)
    SCE_TEX_RESEARCHFOR(".mdl", IL_MDL)
    SCE_TEX_RESEARCHFOR(".mng", IL_MNG)
    SCE_TEX_RESEARCHFOR(".pcx", IL_PCX)
    SCE_TEX_RESEARCHFOR(".pix", IL_PIX)
    SCE_TEX_RESEARCHFOR(".png", IL_PNG)
    SCE_TEX_RESEARCHFOR(".pbm", IL_PNM)
    SCE_TEX_RESEARCHFOR(".pgm", IL_PNM)
    SCE_TEX_RESEARCHFOR(".ppm", IL_PNM)
    SCE_TEX_RESEARCHFOR(".psd", IL_PSD)
    SCE_TEX_RESEARCHFOR(".pxr", IL_PXR)
    SCE_TEX_RESEARCHFOR(".bw", IL_SGI)
    SCE_TEX_RESEARCHFOR(".rgb", IL_SGI)
    SCE_TEX_RESEARCHFOR(".rgba", IL_SGI)
    SCE_TEX_RESEARCHFOR(".sgi", IL_SGI)
    SCE_TEX_RESEARCHFOR(".tga", IL_TGA)
    SCE_TEX_RESEARCHFOR(".tif", IL_TIF)
    SCE_TEX_RESEARCHFOR(".tiff", IL_TIF)
    SCE_TEX_RESEARCHFOR(".wal", IL_WAL)
    SCE_TEX_RESEARCHFOR(".xpm", IL_XPM)
    {
        type = IL_TYPE_UNKNOWN;
    }
    
    #undef SCE_TEX_RESEARCHFOR
    
    return (type);
}
#endif


void SCE_Texture_SetResizeFilter(SCEenum filter)
{
    sce_tex_filter = filter;
}

void SCE_Texture_ReduceImages(int reduce)
{
    sce_tex_reduce = reduce;
}


void* SCE_Texture_DevILLoader(FILE *fp, const char *fname, void *unused)
{
    SCE_STexData *data = NULL;  /* donnees de la texture */
    
    
    /* inutilises */
    fp = NULL;
    unused = NULL;
    
    data = SCE_Texture_CreateData();
    if(data == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    ilGenImages(1, &data->id);
    
    /* binding de l'id devil */
    ilBindImage(data->id);
    
    /* chargement de l'image */
    if(!ilLoadImage(fname))
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("DevIL can't load '%s'", fname);
        ilBindImage(0);
        return (NULL);
    }
    
    /* recuperation des informations de l'image */
    data->cdata->level = 0;
    data->cdata->w = ilGetInteger(IL_IMAGE_WIDTH);
    data->cdata->h = ilGetInteger(IL_IMAGE_HEIGHT);
    data->cdata->internal = data->cdata->format = ilGetInteger(IL_IMAGE_FORMAT);
    data->cdata->type = ilGetInteger(IL_IMAGE_TYPE);
    data->cdata->texels = ilGetData();
    
    /* verification des dimensions de l'image */
    if(!SCE_CHasCap(SCE_TEX_NON_POWER_OF_TWO) &&
       (!SCE_Math_PowerOfTwo(data->cdata->w) ||
        !SCE_Math_PowerOfTwo(data->cdata->h)))
    {
        /* non support des dimensions non puissance de 2, et c'est le cas...
           redimentionnement de l'image via DevIL */
        
        /* recherche des dimensions puissance de 2 les plus proches de w et h */
        /** NOTE: algorithme a revoir d'URGENCE!!! */
        if(sce_tex_reduce)
        {
            while(!SCE_Math_PowerOfTwo(data->cdata->w))
                data->cdata->w--;
            
            while(!SCE_Math_PowerOfTwo(data->cdata->h))
                data->cdata->h--;
        }
        else
        {
            while(!SCE_Math_PowerOfTwo(data->cdata->w))
                data->cdata->w++;
            
            while(!SCE_Math_PowerOfTwo(data->cdata->h))
                data->cdata->h++;
        }
        
        /* application du filtre */
        iluImageParameter(ILU_FILTER, sce_tex_filter);
        
        /* redimensionnement */
        iluScale(data->cdata->w, data->cdata->h, 1);
        
        /* re-recuperation de l'adresse de l'image */
        data->cdata->texels = ilGetData();
        
        /* logization de l'operation */
        #ifdef SCE_LANG_FR
        Logger_PrintMsg("_\nvotre materiel ne supporte pas les textures aux "
                        "dimensions non puissance de 2\n%s de '%s' "
                        " force\nnouvelles dimensions: %d*%d\n", sce_tex_reduce ?
                        "retrecissement":"agrandissement", fname,
                        data->cdata->w, data->cdata->h);
        #else
        /** TODO: traduction a effectuer ici */
        Logger_PrintMsg("_\nvotre materiel ne supporte pas les textures aux "
                        "dimensions non puissance de 2\n%s de '%s' "
                        " force\nnouvelles dimensions: %d*%d\n", sce_tex_reduce ?
                        "retrecissement":"agrandissement", fname,
                        data->cdata->w, data->cdata->h);
        #endif
    }
    
    ilBindImage(0);
    
    return (data);
}

void* SCE_Texture_DDSLoader(FILE *fp, const char *fname, void *unused)
{
    SCE_STexData *data = NULL;      /* donnees de la texture */
    SCE_SDDSImage *dds_tex = NULL;  /* informations de la texture .dds */
    size_t offset = 0;              /* pour parcourir les niveaux de mipmap */
    size_t pxsize = 0;              /* taille d'un pixel (bytes) */
    unsigned int i;
    
    
    /* inutilise */
    unused = NULL;
    
    data = SCE_Texture_CreateData();
    if(data == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    /* chargement de l'image */
    dds_tex = SCE_DDS_LoadFromFile(fp, fname);
    if(dds_tex == NULL)
    {
        Logger_LogSrc();
        SCE_Texture_DeleteData(data);
        return (NULL);
    }
    
    data->n_data = dds_tex->n_levels;
    data->cdata = SCE_realloc(data->cdata, data->n_data * sizeof *data->cdata);
    if(data->cdata == NULL)
    {
        SCE_free(dds_tex->texels);
        SCE_free(dds_tex);
        SCE_Texture_DeleteData(data);
        Logger_LogSrc();
        return (NULL);
    }
    
    
    /* recuperation des informations de l'image */
    data->cdata[0].level = 0;
    data->cdata[0].w = dds_tex->width;
    data->cdata[0].h = dds_tex->height;
    data->cdata[0].internal = data->cdata[0].format = dds_tex->format;
    data->cdata[0].type = /* OSEF si c'est du compresse... */ 0;
    data->cdata[0].texels = dds_tex->texels;
    
    SCE_free(dds_tex);
    data->freetexels = 1; /* attention, ici c'est important */
    
    /* verification des dimensions de l'image */
    if(!SCE_CHasCap(SCE_TEX_NON_POWER_OF_TWO) &&
       (!SCE_Math_PowerOfTwo(data->cdata[0].w) ||
        !SCE_Math_PowerOfTwo(data->cdata[0].h)))
    {
        /* non support des dimensions non puissance de 2, et c'est le cas... */
        Logger_Log(SCE_INVALID_OPERATION);
        #ifdef SCE_LANG_FR
        Logger_LogMsg("votre materiel ne supporte pas les textures aux "
                      "dimensions non puissance de 2, et je ne supporte pas le "
                      "redimensionnement des textures au format compresse :/ "
                      "vous devez redimensionner votre texture");
        #else
        /** TODO: a traduire */
        Logger_LogMsg("your hardware does not support non-power of "
                      "two textures, you must rescale your texture");
        #endif
        
        SCE_Texture_DeleteData(data);
        
        return (NULL);
    }
    
    pxsize = data->cdata[0].pxsize =
    (data->cdata[0].internal == GL_COMPRESSED_RGBA_S3TC_DXT1_EXT) ? 8 : 16;
    
    for(i=1; i<data->n_data; i++)
    {
        /* calcul de l'offset */
        offset =
        ((data->cdata[i-1].w+3)/4) * ((data->cdata[i-1].h+3)/4) * pxsize;
        
        /* copie des donnees globales */
        memcpy(&data->cdata[i], &data->cdata[i-1], sizeof *data->cdata);
        
        /* calcul des donnees membres (propres au <i>eme niveau de mipmap) */
        data->cdata[i].level = i;
        data->cdata[i].w /= 2;
        data->cdata[i].h /= 2;
        data->cdata[i].texels += offset;
        
        /* NOTE: flip(texels, w, h, pixels_size); !! */
    }
    
    /* NOTE: petite merdouille qu'il faudra gerer: le flip de l'image
             pour le loader DevIL, c'est cool: DevIL s'en charge, mais ici,
             il va falloire mettre la main a la pate... */
    
    return (data);
}


int SCE_Texture_AddImage(SCE_STexture *tex, SCEenum type, SCE_STexData *data)
{
    unsigned int id;
    unsigned int i;
    
    #ifdef __SCE_DEBUG__
    if(tex == NULL || data == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    id = tex->n_data;
    tex->n_data++;
    
    tex->data = SCE_realloc(tex->data, tex->n_data * sizeof *tex->data);
    if(tex->data == NULL)
    {
        Logger_LogSrc();
        tex->n_data--;
        return (SCE_ERROR);
    }
    
    tex->data[id] = data;
    
    /* pour chaque sous-image (mipmap) */
    for(i=0; i<tex->data[id]->n_data; i++)
    {
        /* on affecte le type de l'image aux donnees */
        tex->data[id]->cdata[i].target = type;
        
        /* on associe l'image a la texture */
        SCE_CAssociateTextureData(tex->tex, &tex->data[id]->cdata[i]);
    }
    
    return (SCE_OK);
}


int SCE_Texture_Build(SCE_STexture *tex, int use_mipmap)
{
    int hw_mipmap = 0;
    unsigned int i, j;
    
    #ifdef __SCE_DEBUG__
    if(tex == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    
    /* envoie des donnees (images) au coeur */
    for(i=0; i<tex->n_data; i++)
    {
        for(j=0; j<tex->data[i]->n_data; j++)
        {
            SCE_CSetTextureData(tex->tex, &tex->data[i]->cdata[j]);
        }
    }
    
    /* par defaut, le moteur tente de faire generer les mipmaps en hardware */
    if(use_mipmap && SCE_CHasCap(SCE_TEX_HW_GEN_MIPMAP))
        hw_mipmap = 1;
    
    /* construction */
    if(SCE_CBuildTexture(tex->tex, use_mipmap, hw_mipmap) < 0)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    /* installation d'un filtre */
    if(use_mipmap || tex->tex->spec_mipmap)
        SCE_CSetTextureFilter(tex->tex, SCE_TEX_TRILINEAR);
    else
        SCE_CSetTextureFilter(tex->tex, SCE_TEX_LINEAR);
    
    return (SCE_OK);
}


int SCE_Texture_Update(SCE_STexture *tex)
{
    #ifdef __SCE_DEBUG__
    if(tex == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    if(SCE_CUpdateTexture(tex->tex) < 0)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    return (SCE_OK);
}


void SCE_Texture_Use(unsigned int unit, SCE_STexture *tex)
{
    /* NOTE: a optimiser... ? (voir SCECTexture) */
    if(tex != NULL)
        SCE_CUseTexture(unit, tex->tex, 1);
    else
        SCE_CUseTexture(unit, NULL, 0);
}
